﻿using System;
using System.Collections.Generic;
using System.Runtime.CompilerServices;
using System.Threading;
using System.Threading.Tasks;

namespace ConsoleApp1
{
    class Program
    {
        static void Main(string[] args)
        {
            AsyncMain();
            Console.ReadLine();
        }

        private static async void AsyncMain()
        {
            await TestAwait();
            Console.WriteLine("test");

            await TaskAwait();
            Console.WriteLine("test");
        }

        private static async MyTask TestAwait()
        {
            MyTask task = new MyTask(() =>
            {
                Console.WriteLine("Before-MyTask-Sleep");
                Thread.Sleep(5000);
                Console.WriteLine("After-MyTask-Sleep");
            });

            await task;
        }

        private static async Task TaskAwait()
        {
            await new TaskFactory().StartNew(() => 
            {
                Console.WriteLine("Before-Task-Sleep");
                Thread.Sleep(5000);
                Console.WriteLine("After-Task-Sleep");
            });
        }

    }

    public struct AsyncMyTaskMethodBuilder
    {
        AsyncTaskMethodBuilder _methodBuilder;

        public MyTask Task
        {
            get
            {
                return new MyTask(_methodBuilder.Task);
            }
        }

        public static AsyncMyTaskMethodBuilder Create()
        {
            return new AsyncMyTaskMethodBuilder { _methodBuilder = AsyncTaskMethodBuilder.Create() };
        }

        public void SetException(Exception exception)
        {
            _methodBuilder.SetException(exception);
        }

        public void SetResult()
        {
            _methodBuilder.SetResult();
        }

        public void AwaitOnCompleted<TAwaiter, TStateMachine>(ref TAwaiter awaiter, ref TStateMachine stateMachine) where TAwaiter : INotifyCompletion where TStateMachine : IAsyncStateMachine
        {
            _methodBuilder.AwaitOnCompleted(ref awaiter, ref stateMachine);
        }

        public void AwaitUnsafeOnCompleted<TAwaiter, TStateMachine>(ref TAwaiter awaiter, ref TStateMachine stateMachine) where TAwaiter : ICriticalNotifyCompletion where TStateMachine : IAsyncStateMachine
        {
            _methodBuilder.AwaitUnsafeOnCompleted(ref awaiter, ref stateMachine);
        }

        public void SetStateMachine(IAsyncStateMachine stateMachine)
        {
            _methodBuilder.SetStateMachine(stateMachine);
        }

        public void Start<TStateMachine>(ref TStateMachine stateMachine) where TStateMachine : IAsyncStateMachine
        {
            _methodBuilder.Start(ref stateMachine);
        }
    }

    [AsyncMethodBuilder(typeof(AsyncMyTaskMethodBuilder))]
    public class MyTask
    {
        Thread _t;
        Task _task;

        internal Task Task
        {
            get { return _task; }
        }

        bool _isCompleted;
        List<Action> _continuation = new List<Action>();

        public void AddContinuation(Action action)
        {
            if (_t == null)
            {
                _task.ContinueWith((t) => { action(); });
            }
            else
            {
                _continuation.Add(action);
            }
        }

        public bool IsCompleted
        {
            get
            {
                if (_t == null)
                {
                    return _task.IsCompleted;
                }
                
                return _isCompleted;
            }
        }

        public MyTask(Task task)
        {
            _task = task;
        }

        public MyTask(Action action)
        {
            _t = new Thread(
                (ThreadStart)( 
                () =>
                {
                    action();
                    _isCompleted = true;

                    foreach (var item in _continuation)
                    {
                        item();
                    }
                }
                )
            );

            _t.Start();
        }

        public MyTaskAwaiter GetAwaiter()
        {
            MyTaskAwaiter ta = new MyTaskAwaiter(this);
            return ta;
        }
    }

    public struct MyTaskAwaiter : ICriticalNotifyCompletion, INotifyCompletion
    {
        MyTask _task;

        public MyTaskAwaiter(MyTask t)
        {
            _task = t;
        }

        public bool IsCompleted
        {
            get
            {
                return _task.IsCompleted;
            }
        }

        public void GetResult()
        {
        }

        public void OnCompleted(Action continuation)
        {
            if (_task.Task == null)
            {
                _task.AddContinuation(continuation);
            }
            else
            {
                _task.Task.GetAwaiter().OnCompleted(continuation);
            }
        }

        public void UnsafeOnCompleted(Action continuation)
        {
            if (_task.Task == null)
            {
                _task.AddContinuation(continuation);
            }
            else
            {
                _task.Task.GetAwaiter().UnsafeOnCompleted(continuation);
            }
        }
    }
}
